'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

exports.default = function (_ref) {
    var t = _ref.types;

    function insertVariableDeclarationBeforePathAndUpdateScope(id, path) {
        var variableDeclaration = t.variableDeclaration('let', [t.variableDeclarator(id)]);
        path.insertBefore(variableDeclaration);

        path.scope.bindings[id.name] = new _binding2.default({
            identifier: id,
            existing: null,
            scope: path.scope,
            path: path,
            kind: 'let'
        });
    }

    var expressionStatementTransformer = {
        Identifier: function Identifier(path) {
            var pathParent = path.parent;
            var pathNode = path.node;

            if (t.isObjectProperty(pathParent) && pathNode === pathParent.value || t.isArrayPattern(pathParent) || t.isAssignmentPattern(pathParent) && pathNode === pathParent.left || t.isRestElement(pathParent) || t.isRestProperty(pathParent)) {

                if (!this.expressionStatementPath.scope.hasBinding(pathNode.name)) {
                    insertVariableDeclarationBeforePathAndUpdateScope(pathNode, this.expressionStatementPath);
                }
            }
        }
    };

    var forStatementTransformer = {
        Identifier: function Identifier(path) {
            var pathParent = path.parent;
            var pathNode = path.node;

            if (this.undeclaredId === undefined) {
                if (t.isObjectProperty(pathParent) && pathNode === pathParent.value || t.isArrayPattern(pathParent) || t.isAssignmentPattern(pathParent) && pathNode === pathParent.left) {

                    if (!this.forStatementPath.scope.hasBinding(pathNode.name)) {
                        var uid = this.forStatementPath.scope.generateUidIdentifier(pathNode.name);
                        insertVariableDeclarationBeforePathAndUpdateScope(uid, this.forStatementPath);
                        this.forStatementPath.traverse(forStatementTransformer, { undeclaredId: pathNode,
                            declaredId: uid });
                    }
                }
            } else if (this.undeclaredId.name === pathNode.name) {
                path.replaceWith(this.declaredId);
            }
        }
    };

    var forOfStatementTransformer = {
        Identifier: function Identifier(path) {
            var pathParent = path.parent;
            var pathNode = path.node;

            if (this.undeclaredId === undefined) {
                if (t.isObjectProperty(pathParent) && pathNode === pathParent.value || t.isArrayPattern(pathParent) || t.isAssignmentPattern(pathParent) && pathNode === pathParent.left) {

                    if (!this.forOfStatementPath.scope.hasBinding(pathNode.name)) {
                        var uid = this.forOfStatementPath.scope.generateUidIdentifier(pathNode.name);
                        insertVariableDeclarationBeforePathAndUpdateScope(uid, this.forOfStatementPath);
                        this.forOfStatementPath.traverse(forOfStatementTransformer, { undeclaredId: pathNode,
                            declaredId: uid });
                    }
                }
            } else if (this.undeclaredId.name === pathNode.name) {
                path.replaceWith(this.declaredId);
            }
        }
    };

    var visitorWrapper = {
        visitor: {
            ExpressionStatement: function ExpressionStatement(path) {
                if (this.lastPath === undefined) {

                    if (t.isAssignmentExpression(path.node.expression)) {
                        var left = path.node.expression.left;
                        if (t.isIdentifier(left)) {
                            if (!path.scope.hasBinding(left.name)) {
                                insertVariableDeclarationBeforePathAndUpdateScope(left, path);
                            }
                        } else {
                            // ArrayPattern, ObjectPattern, etc.
                            var subPath = path.get('expression.left');
                            subPath.traverse(expressionStatementTransformer, { expressionStatementPath: path });
                        }
                        // We have to do this in order for the sibling nodes to see the changes
                        path.parentPath.traverse(visitorWrapper.visitor, { lastPath: path });
                        path.stop();
                    }
                } else if (this.lastPath === path) {
                    delete this.lastPath;
                }
            },
            ForStatement: function ForStatement(path) {
                if (this.lastPath === undefined) {
                    var init = path.node.init;
                    if (t.isAssignmentExpression(init)) {
                        if (t.isIdentifier(init.left)) {
                            if (!path.scope.hasBinding(init.left.name)) {
                                var newInit = t.variableDeclaration('let', [t.variableDeclarator(init.left, init.right)]);
                                var newPath = t.forStatement(newInit, path.node.test, path.node.update, path.node.body);
                                path.replaceWith(newPath);
                            }
                        } else {
                            // ArrayPattern, ObjectPattern, etc.
                            var subPath = path.get('init.left');
                            subPath.traverse(forStatementTransformer, { forStatementPath: path });
                        }
                    } else if (t.isSequenceExpression(init)) {
                        for (var i = 0; i < init.expressions.length; i += 1) {
                            var expression = init.expressions[i];

                            if (t.isIdentifier(expression.left)) {
                                if (!path.scope.hasBinding(expression.left.name)) {
                                    var id = expression.left;
                                    var uid = path.scope.generateUidIdentifier(id.name);
                                    insertVariableDeclarationBeforePathAndUpdateScope(uid, path);
                                    path.traverse(forStatementTransformer, { undeclaredId: id,
                                        declaredId: uid });
                                }
                            } else {
                                // ArrayPattern, ObjectPattern, etc.
                                var _subPath = path.get('init.expressions.' + i + '.left');
                                _subPath.traverse(forStatementTransformer, { forStatementPath: path });
                            }
                        }
                    }
                }
            },
            ForOfStatement: function ForOfStatement(path) {
                if (this.lastPath === undefined) {
                    var left = path.node.left;

                    if (t.isIdentifier(left)) {
                        if (!path.scope.hasBinding(left.name)) {
                            var newLeft = t.variableDeclaration('let', [t.variableDeclarator(left)]);
                            var newPath = t.forOfStatement(newLeft, path.node.right, path.node.body);
                            path.replaceWith(newPath);
                        }
                    } else {
                        // ArrayPattern, ObjectPattern, etc.
                        var subPath = path.get('left');
                        subPath.traverse(forOfStatementTransformer, { forOfStatementPath: path });
                    }
                }
            },
            ForInStatement: function ForInStatement(path) {
                if (this.lastPath === undefined) {
                    var left = path.node.left;
                    if (t.isIdentifier(left) && !path.scope.hasBinding(left.name)) {
                        var newLeft = t.variableDeclaration('let', [t.variableDeclarator(left)]);
                        var newPath = t.forInStatement(newLeft, path.node.right, path.node.body);
                        path.replaceWith(newPath);
                    }
                }
            }
        }
    };
    return visitorWrapper;
};

var _binding = require('babel-traverse/lib/scope/binding');

var _binding2 = _interopRequireDefault(_binding);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }